<?php
namespace HCM\Util\ModeloImportacao;

use HCM\Util\ModeloImportacao\References\Autoincrement;
use HCM\Util\ModeloImportacao\References\Executavel;
use HCM\Util\ModeloImportacao\References\FixedValue;
use HCM\Util\ModeloImportacao\References\ForeignKey;
use HCM\Util\ModeloImportacao\References\Label;
use HCM\Util\ModeloImportacao\References\PadraoInvalido;
use HCM\Util\ModeloImportacao\References\Reference;
use HCM\Util\ModeloImportacao\References\ValorCampo;
use HCM\Util\ModeloImportacao\References\FixedField;

/**
 * Constroi o conjunto de dados que sera utilizado para a criacao das queries da importacao.
 *
 * Class DataBuilder
 * @package Importacao\Logic\Bases\ModeloImportacao
 */
class DataBuilder
{
	private $agrupamento;
	private $data;
	private $agrupadores;
	private $rules;
	private $fixedFields;
	private $padroesMetadados;
	private $entidades;
    private $excludedTables;

    static $pk = array();

    /**
     *
     *
     * @param $agrupamento array Relacao entre os dados dos arquivos excel e as tabelas do banco de dados (esturura.json).
     * @param $data        array Array com todos os dados lidos dos arquivos de excel.
     * @param $helpers     array Array com todos os dados auxiliares do modelo de importacao (helper.json).
     */
    public function __construct($agrupamento, $data, $helpers)
	{
		$this->agrupamento     = $agrupamento;
		$this->data            = $data;
		$this->rules           = $helpers["rules"];
        $this->fixedFields     = $helpers["fixedFields"];
        // Armazena o nome dos agrupadores (presentes no arquivo estrutura.json) que serão utilizados em determinado projeto de importacao.
        $this->agrupadores     = array();
        // Objeto onde todas as entidades criadas serao planificadas. Este array sera retornado no final do processo de geracao dos dados.
        // É publicamente acessivel pelo metodo addAgrupador.
        $this->entidades       = array();

        $this->excludedTables = $helpers["excludedTables"];

        // Este array armazena todos os padroes que sao compreendidos pelo modelo, e sao utilizados nos arquivos de estrutura.json e helper.json. Qualquer classe
        // utilizada neste array deverá implementar a interface de HCM\Util\ModeloImportacao\References\Reference.
		$this->padroesMetadados = array(
			new ValorCampo(),
			new ForeignKey(),
            new Label(),
            new FixedField(),
            new FixedValue(),
            new Executavel(),
			new PadraoInvalido(),
		);
	}

	public function addAgrupador($agrupador) 
	{
		$this->agrupadores[] = $agrupador;
	}

	public function build() 
	{
		$this->percorreLinhasDados();
		// Retorna todos os dados planos e insere todos os campos padrao.
		return $this->entidades;
	}

    /**
     * Percorre todas as linhas do excel.
     */
    protected function percorreLinhasDados()
	{
        // $progressBar = new \Importacao\Logic\Util\BarraProgressoCli(
        //     'Criando Arquivo SQL', count($this->data)
        // );

		foreach ($this->data as $key => $linhaDados) {
            // echo ($key - 1) . "\n";
            $this->percorreAgrupamentos($linhaDados);
            // $progressBar->showStatus($key + 1);
            unset($this->data[$key]);
		}

        // $progressBar->finish();
	}

    /**
     * Percorre todos os agrupamentos do arquivo estrutura.json
     *
     * @param $linhaDados array objeto contendo a linha corrente do excel. Passa por este metodo somente para ser transmitido para os proximos,
     * onde este realmente sera necessario.
     */
    protected function percorreAgrupamentos($linhaDados)
	{
		foreach ($this->agrupamento as $nomeAgrupamento => $tabelas) {
            if (in_array($nomeAgrupamento, $this->agrupadores)) {
                $entityContainer = new EntityContainer();
                $this->percorreTabelas($linhaDados, $tabelas, $entityContainer);
                if (isset($this->rules[$nomeAgrupamento])) {
                    $this->executaRegrasDeTabelas($this->rules[$nomeAgrupamento], $entityContainer, $linhaDados);
                }

                $this->impedeRepetirPk($entityContainer);
            }
		}
	}

    /**
     * Percorre todas as regras descritas para determinada tabela (se estas existirem). Apos executar todas as regras desta tabela,
     * este metodo tambem faz chamada ao metodo que executa as regras para os campos destas tabelas.
     *
     * As regras sao descritas no arquivo helper.json, sendo que cada regra deve conter as chaves de "condicao", onde é descrita
     * a condicao a ser verificada, "entao", onde e descrito o que deve ser feito, caso a condicao seja satisfeita, e "senao", que
     * descreve o que deve ser feito caso a condicao nao seja satisfeita.
     *
     * Importante notar que, sempre que a condicao resultar em "false", a entidade (tabela) em questao sera removida do result.
     *
     * @param $regras                          array Array contendo o conjunto condica, entao, senao
     * @param EntityContainer $entityContainer       Container armazenando todas as entidades de determinada linha.
     * @param $linhaDados                      array Array contendo os dados da linha do arquivo excel.
     */
    protected function executaRegrasDeTabelas($regras, EntityContainer $entityContainer, $linhaDados)
	{
		foreach ($regras["tabelas"] as $nomeTabela => $tabela) {
			foreach ($tabela["regras"] as $regra) {
				$condicao = $this->verificaCondicaoRegra(
					$regra["condicao"], $nomeTabela, $entityContainer, $linhaDados
				);
				if ($condicao) {
					$this->getDadosReference(
						$nomeTabela, null, $regra["entao"], null, $entityContainer, $linhaDados
					);
				} else {
					$entityContainer->removeTabela($nomeTabela);
					if (isset($regra["senao"])) {
						$this->getDadosReference(
							$nomeTabela, null, $regra["senao"], null, $entityContainer, $linhaDados
						);
					}
				}
			}

			if (isset($tabela["colunas"])) {
				$this->executaRegrasDeColuna($tabela["colunas"], $nomeTabela, $entityContainer, $linhaDados);
			}
		}
	}

    /**
     * Executa as regras a nivel de coluna. Assim como nas regras de tabela, estas devem possuir condicao, entao e senao.
     *
     * @param $regrasColuna
     * @param $nomeTabela
     * @param EntityContainer $entityContainer
     * @param $linhaDados
     * @return bool
     */
    protected function executaRegrasDeColuna($regrasColuna, $nomeTabela, EntityContainer $entityContainer, $linhaDados)
	{
        /** @var  $entidade BasicEntity*/
        $entidade = $entityContainer->getTabela($nomeTabela);
        if (!$entidade) {
            return false;
        }

        foreach ($regrasColuna as $colunaName => $coluna) {
            foreach ($coluna['regras'] as $regra) {
                $condicao = $this->verificaCondicaoRegra(
                    $regra["condicao"], $nomeTabela, $entityContainer, $linhaDados
                );

                if ($condicao) {
                    $result = $this->getDadosReference(
                        $nomeTabela, null, $regra["entao"], null, $entityContainer, $linhaDados
                    );
                    $entidade->addField($colunaName, $result);
                } else if (isset($regra["senao"])) {
					$result = $this->getDadosReference(
						$nomeTabela, null, $regra["senao"], null, $entityContainer, $linhaDados
					);
					$entidade->addField($colunaName, $result);
				}
			}
		}
        return true;
	}

    /**
     * Metodo utilizado para resolucao da condicao.
     *
     * @param $regra          string            Padrao informado no helper.json, que apos processado, possui a condicao necessaria.
     * @param $nomeTabela     string            Nome da tabela sobre a qual a condicao sera imposta
     * @param EntityContainer $entityContainer  Container com todas as entidades de determinada linha.
     * @param $linhaDados     array             Linha de dados do excel
     *
     * @return boolean
     */
    protected function verificaCondicaoRegra($regra, $nomeTabela, EntityContainer $entityContainer, $linhaDados)
	{
		$condicao = $this->getDadosReference($nomeTabela, null, $regra, null, $entityContainer, $linhaDados);
        $result = null;
		eval('$result = ('.$condicao.');');
		return $result;
	}

    /**
     * Percorre todas as tabelas de determinado agrupamento do arquivo de estruturas (estrutura.json).
     *
     * @param $linhaDados     array             Linha do excel. Nao e utilizado por este metodo, somente transmitido para os
     *                                          metodos posteriores.
     * @param $tabelas        array             Array com todas as tabelas do agrupamento corrente.
     * @param EntityContainer $entityContainer  Container que devera armazenar todas as entidades geradas a partir da linha.
     */
    protected function percorreTabelas($linhaDados, $tabelas, EntityContainer $entityContainer) {
        foreach ($tabelas as $nomeTabela => $dadosTabela) {
            if (!in_array($nomeTabela, $this->excludedTables)) {
    			$entityContainer->adicionaTabela($nomeTabela, new BasicEntity($nomeTabela));
    			$this->percorreDados($linhaDados, $nomeTabela, $dadosTabela, $entityContainer);
    			$this->insereCamposPadrao($entityContainer->getTabela($nomeTabela));
            }
		}
	}

    /**
     * Percorre todos os dados de uma linha do excel.
     *
     * @param $linhaDados     array            Linha do excel.
     * @param $nomeTabela     string           Nome da tabela.
     * @param $dadosTabela    array            conteudo das tabelas descritas em estrutura.json. Nao e utilizado aqui, mas
     *                                         transmitido para os metodos posteriores.
     * @param EntityContainer $entityContainer Container que devera armazenar todas as entidades geradas a partir da linha.
     */
    protected function percorreDados($linhaDados, $nomeTabela, $dadosTabela, EntityContainer $entityContainer)
	{
        $this->addPkTabela($nomeTabela, $dadosTabela, $entityContainer);
        foreach ($linhaDados as $nomeColunaDados => $valorColunaDados) {
			$this->percorreDadosAgrupamento ( 
				$nomeTabela, $dadosTabela, $nomeColunaDados, $valorColunaDados, $linhaDados, $entityContainer
			);
		}
        $this->addPkValor($nomeTabela, $entityContainer);
	}

    /**
     * Percorre todos os dados de uma tabela, dentro de um agrupamento.
     *
     * @param $nomeTabela       string
     * @param $dadosTabela
     * @param $nomeColunaDados
     * @param $valorColunaDados 
     * @param $linhaDados
     * @param EntityContainer $entityContainer
     */
    protected function percorreDadosAgrupamento($nomeTabela, $dadosTabela, $nomeColunaDados,
											  $valorColunaDados, $linhaDados, EntityContainer $entityContainer)
	{
		foreach ($dadosTabela as $nomeColuna => $dadosColuna) {
			$this->constroiFieldEntidade(
				$nomeTabela, $nomeColunaDados, $valorColunaDados, $nomeColuna, 
				$dadosColuna, $linhaDados, $entityContainer
			);
		}
	}

	protected function constroiFieldEntidade($nomeTabela, $nomeColunaDados, $valorColunaDados, 
		                                   $nomeColuna, $dadosColuna, $linhaDados, EntityContainer $entityContainer)
	{
        /** @var  $entidade BasicEntity*/
		$entidade        = $entityContainer->getTabela($nomeTabela);
		// $nomeColuna      = strtoupper(Util::removeCaracteresInvalidos($nomeColuna));
		// $nomeColunaDados = strtoupper(Util::removeCaracteresInvalidos($nomeColunaDados));

		if ($nomeColuna === $nomeColunaDados) {
			$this->addFields($entidade, $dadosColuna, $valorColunaDados);
		} else if (preg_match("/\#/", $nomeColuna) && !$entidade->hasColumn($nomeColuna)) {
			$this->adicionaCamposReference(
				$nomeTabela, $nomeColuna, $dadosColuna, $valorColunaDados, $entidade, $linhaDados, $entityContainer
			);
		}
	}

	protected function adicionaCamposReference($nomeTabela, $nomeColuna, $dadosColuna, $valorColunaDados, 
											 BasicEntity $entidade, $linhaDados, EntityContainer $entityContainer)
	{
		$dadoColuna = $this->getDadosReference (
			$nomeTabela, $nomeColuna, $dadosColuna, $valorColunaDados, $entityContainer, $linhaDados
		);

		if (!preg_match("/".Reference::PADRAO_INVALIDO."/", $dadoColuna)) {
			$entidade->addField(str_replace("#", "", $nomeColuna), $dadoColuna);
		}
	}

	protected function addFields(BasicEntity $entidade, $dadosColuna, $valorColunaDados)
	{
		$fields = explode(",", str_replace(" ", "", $dadosColuna));
		foreach ($fields as $field) {
			$entidade->addField($field, $valorColunaDados);
		}
	}

	// Retorna o valor esperado para uma coluna informada nos agrupamentos através de meta-dados.
	protected function getDadosReference($nomeTabela, $nomeColuna, $metaDado, $valorFixo,
                                       EntityContainer $entityContainer, $linhaDados = null)
	{
		$padraoEncontrado = false;
        $matches = array();
        /** @var Reference $padrao */
        foreach ($this->padroesMetadados as $padrao) {
			if (preg_match_all($padrao->getRegularExpression(), $metaDado, $matches)) {
                /* Itera a ocorrencia de varios padroes juntos */
                for ($i = count($matches[0]); $i > 0; $i--) {
                    $metaDado = $padrao->getValorReferencia(
                        $nomeTabela, $nomeColuna, $metaDado, $valorFixo, $entityContainer, $linhaDados, $this->fixedFields
                    );
                }
				$padraoEncontrado = true;
			}
		}
        
		return !$padraoEncontrado ? $metaDado : $this->getDadosReference (
			$nomeTabela, $nomeColuna, $metaDado, $valorFixo, $entityContainer, $linhaDados
		);
	}

	protected function insereCamposPadrao(BasicEntity $entidade) 
	{
		foreach ($this->fixedFields as $name => $value) {
			if (!$entidade->hasColumn($name)) {
				$entidade->addField($name, $value);
			}
		}
	}

    /**
     * @param $nomeTabela
     * @param $dadosTabela
     * @param EntityContainer $entityContainer
     */
    protected function addPkTabela($nomeTabela, $dadosTabela, EntityContainer $entityContainer)
    {
        if (isset($dadosTabela['PK'])) {
            $entityContainer->getTabela($nomeTabela)->setPk($dadosTabela['PK']);
            unset($dadosTabela['PK']);
        }
    }

    /**
     * @param $nomeTabela
     * @param EntityContainer $entityContainer
     */
    protected function addPkValor($nomeTabela, EntityContainer $entityContainer)
    {
        $pk = $entityContainer->getTabela($nomeTabela)->getPk();
        if (!$pk) return;
        foreach ($pk as $key => $value) {
            $pk[$key] = $entityContainer->getTabela($nomeTabela)->getField($key);
        }
        $entityContainer->getTabela($nomeTabela)->setPk($pk);
    }

    /**
     * @param EntityContainer $entityContainer
     */
    protected function impedeRepetirPk(EntityContainer $entityContainer)
    {
        foreach ($entityContainer->getTabelas() as $tabela) {
            $repete = false;
            if ($tabela->getPk()) {
                if (isset(self::$pk[$tabela->getTableName()])) {
                    foreach (self::$pk[$tabela->getTableName()] as $pk) {
                        if ($pk && ($pk == $tabela->getPk())) {
                            $repete = true;
                            break;
                        }
                    }
                }
                if (!$repete) {
                    self::$pk[$tabela->getTableName()][] = $tabela->getPk();
                }
            }
            if (!$repete) {
                array_push($this->entidades, $tabela);
            }
        }
    }
}